@@ -32,7 +32,7 @@ class Future(I):
 
     """
 
-    num_ops = 1
+    num_ops = 2
 
 
 # This is a bit complex. Instead of a "user_future_id", just provide
@@ -177,6 +177,14 @@ class LessThan(I):
     pass
 
 
+class OpAnd(I):
+    pass
+
+
+class OpOr(I):
+    pass
+
+
 class Atomp(I):
     """Check whether something is an atom"""
 
@@ -260,3 +268,7 @@ class Print(I):
 
 class Sleep(I):
     """Sleep for some time"""
+
+
+class GetSessionId(I):
+    """Get the current session ID"""
