@@ -4,6 +4,8 @@ import logging
 from functools import singledispatch, singledispatchmethod, wraps
 from typing import Dict, Tuple
 
+from ..cli.interface import format_source_problem
+from ..exceptions import UserResolvableError
 from ..machine import instructionset as mi
 from ..machine import types as mt
 from ..machine.executable import Executable
@@ -23,8 +25,17 @@ def flatten(list_of_lists: list) -> list:
     return list(itertools.chain.from_iterable(list_of_lists))
 
 
-class CompileError(Exception):
-    pass
+class TealCompileError(UserResolvableError):
+    def __init__(self, node: nodes.Node, msg):
+        if not isinstance(node, nodes.Node):
+            raise TypeError(node)
+        explanation = format_source_problem(
+            node.source_filename,
+            node.source_lineno,
+            node.source_line,
+            node.source_column,
+        )
+        super().__init__(msg, explanation)
 
 
 def optimise_block(n: nodes.N_Definition, block: nodes.N_Progn):
@@ -37,15 +48,15 @@ def optimise_block(n: nodes.N_Definition, block: nodes.N_Progn):
     if isinstance(last, nodes.N_Call) and last.fn.name == n.name:
         # recursive call, optimise it! Replace the N_Call with direct evaluation
         # of the arguments and a jump back to the start
-        new_last_items = list(last.args) + [nodes.N_Goto(None, START_LABEL)]
-        return_values = nodes.N_MultipleValues(None, new_last_items)
-        return nodes.N_Progn(None, block.exprs[:-1] + [return_values])
+        goto_start = list(last.args) + [nodes.N_Goto.from_node(last, START_LABEL)]
+        return_values = nodes.N_MultipleValues(None, None, None, None, goto_start)
+        return nodes.N_Progn.from_node(block, block.exprs[:-1] + [return_values])
 
     elif isinstance(last, nodes.N_If):
-        new_cond = nodes.N_If(
-            None, last.cond, optimise_block(n, last.then), optimise_block(n, last.els)
+        new_cond = nodes.N_If.from_node(
+            last, last.cond, optimise_block(n, last.then), optimise_block(n, last.els)
         )
-        return nodes.N_Progn(None, block.exprs[:-1] + [new_cond])
+        return nodes.N_Progn.from_node(block, block.exprs[:-1] + [new_cond])
 
     else:
         # Nothing to optimise
@@ -76,7 +87,7 @@ def replace_gotos(code: list):
             # + 1 to compensate for the fact that the IP is advanced before
             # the current instruction is evaluated.
             offset = labels[instr.name] - (idx + 1)
-            code[idx] = mi.Jump(mt.TlInt(offset))
+            code[idx] = mi.Jump.from_node(instr, mt.TlInt(offset))
 
     return code
 
@@ -96,7 +107,7 @@ class CompileToplevel:
         """Make a new executable function object with a unique name, and save it"""
         count = len(self.functions)
         identifier = f"#{count}:{name}"
-        start_label = nodes.N_Label(None, START_LABEL)
+        start_label = nodes.N_Label.from_node(n, START_LABEL)
         code = self.compile_function(optimise_tailcall(n))
         fn_code = replace_gotos([start_label] + code)
         self.functions[identifier] = fn_code
@@ -106,42 +117,45 @@ class CompileToplevel:
     def compile_function(self, n: nodes.N_Definition) -> list:
         """Compile a function into executable code"""
         bindings = flatten(
-            [[mi.Bind(mt.TlSymbol(arg)), mi.Pop()] for arg in reversed(n.paramlist)]
+            [
+                [mi.Bind.from_node(n, mt.TlSymbol(arg)), mi.Pop.from_node(n)]
+                for arg in reversed(n.paramlist)
+            ]
         )
         body = self.compile_expr(n.body)
-        return bindings + body + [mi.Return()]
+        return bindings + body + [mi.Return.from_node(n)]
 
     ## At the toplevel, no executable code is created - only bindings
 
     @singledispatchmethod
     def compile_toplevel(self, n) -> None:
-        raise CompileError(f"{n}: Invalid at top level")
+        raise TealCompileError(n, f"Invalid expression type at top level: {type(n)}")
 
     @compile_toplevel.register
     def _(self, n: nodes.N_Call):
         if not isinstance(n.fn, nodes.N_Id) or n.fn.name != "import":
-            raise CompileError(f"{n}: Only `import' can be called at top level")
+            raise TealCompileError(n, f"Only `import' can be called at top level")
 
         if len(n.args) not in (2, 3):
-            raise CompileError(f"{n}: usage: import(name, source, [qualifier])")
+            raise TealCompileError(n, f"usage: import(name, source, [qualifier])")
 
         if not isinstance(n.args[0].value, nodes.N_Id):
-            raise CompileError(f"{n}: Import name must be an identifier")
+            raise TealCompileError(n, f"Import name must be an identifier")
 
         if not isinstance(n.args[1].value, nodes.N_Id):
-            raise CompileError(f"{n}: Import source must be an identifier")
+            raise TealCompileError(n, f"Import source must be an identifier")
 
         import_symb = n.args[0].value.name
         from_kw = n.args[1].symbol
         from_val = n.args[1].value.name
 
         if from_kw and from_kw.name != ":python":
-            raise CompileError(f"{n}: Can't import non-python")
+            raise TealCompileError(n, f"Can't import non-python")
 
         if len(n.args) == 3:
             # TODO? check n.args[2].symbol == ":as"
             if not isinstance(n.args[2].value, nodes.N_Id):
-                raise CompileError(f"{n}: Import qualifier must be an identifier")
+                raise TealCompileError(n, f"Import qualifier must be an identifier")
 
             as_val = n.args[2].value.name
         else:
@@ -177,21 +191,23 @@ class CompileToplevel:
         # TODO?! closures! Need a mi.MakeClosure instruction that updates the
         # top value (TlFunctionPtr) on the stack to reference the current
         # activation record. The Call logic would be different too.
-        return [mi.PushV(mt.TlFunctionPtr(identifier, stack))]
+        return [mi.PushV.from_node(n, mt.TlFunctionPtr(identifier, stack))]
 
     @compile_expr.register
     def _(self, n: nodes.N_Literal):
         val = mt.to_teal_type(n.value)
-        return [mi.PushV(val)]
+        return [mi.PushV.from_node(n, val)]
 
     @compile_expr.register
     def _(self, n: nodes.N_Id):
-        return [mi.PushB(mt.TlSymbol(n.name))]
+        return [mi.PushB.from_node(n, mt.TlSymbol(n.name))]
 
     @compile_expr.register
     def _(self, n: nodes.N_Progn):
         # only keep the last result
-        discarded = flatten(self.compile_expr(exp) + [mi.Pop()] for exp in n.exprs[:-1])
+        discarded = flatten(
+            self.compile_expr(exp) + [mi.Pop.from_node(n)] for exp in n.exprs[:-1]
+        )
         return discarded + self.compile_expr(n.exprs[-1])
 
     @compile_expr.register
@@ -207,7 +223,7 @@ class CompileToplevel:
         return (
             arg_code
             + self.compile_expr(n.fn)
-            + [instr(mt.TlInt(len(n.args)), source=n.index)]
+            + [instr.from_node(n, mt.TlInt(len(n.args)))]
         )
 
     @compile_expr.register
@@ -230,7 +246,7 @@ class CompileToplevel:
         if not isinstance(n.expr, (nodes.N_Call, nodes.N_Id)):
             raise ValueError(f"Can't use await with {n.expr} - {type(n.expr)}")
         val = self.compile_expr(n.expr)
-        return val + [mi.Wait(mt.TlInt(0))]
+        return val + [mi.Wait.from_node(n, mt.TlInt(0))]
 
     @compile_expr.register
     def _(self, n: nodes.N_If):
@@ -240,9 +256,9 @@ class CompileToplevel:
         return [
             # --
             *cond_code,
-            mi.JumpIf(mt.TlInt(len(else_code) + 1)),  # to then_code
+            mi.JumpIf.from_node(n, mt.TlInt(len(else_code) + 1)),  # to then_code
             *else_code,
-            mi.Jump(mt.TlInt(len(then_code))),  # to Return
+            mi.Jump.from_node(n, mt.TlInt(len(then_code))),  # to Return
             *then_code,
         ]
 
@@ -253,12 +269,19 @@ class CompileToplevel:
         if n.op == "=":
             if not isinstance(n.lhs, nodes.N_Id):
                 raise ValueError(f"Can't assign to non-identifier {n.lhs}")
-            return rhs + [mi.Bind(mt.TlSymbol(str(n.lhs.name)))]
+            return rhs + [mi.Bind.from_node(n, mt.TlSymbol(str(n.lhs.name)))]
 
         else:
             lhs = self.compile_expr(n.lhs)
             # TODO check arg order. Reverse?
-            return rhs + lhs + [mi.PushB(mt.TlSymbol(str(n.op))), mi.Call(mt.TlInt(2))]
+            return (
+                rhs
+                + lhs
+                + [
+                    mi.PushB.from_node(n, mt.TlSymbol(str(n.op))),
+                    mi.Call.from_node(n, mt.TlInt(2)),
+                ]
+            )
 
 
 ###
